<?php

namespace ObsMania;


use ManiaControl\Settings\SettingManager;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\TimerListner;
use ManiaControl\Commands\CommandListener;
use ManiaControl\ManiaControl;
use ManiaControl\Manialinks\ManialinkPageAnswerListener;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Maps\MapManager;
use ManiaControl\Plugins\Plugin;
use ManiaControl\Callbacks\Models\RecordCallback;
use ManiaControl\Manialinks\ManialinkManager;
use ManiaControl\Utils\Formatter;
use ManiaControl\Callbacks\TimerListener;
use ManiaControl\Admin\AuthenticationManager;
use FML\Controls\Control;
use FML\Controls\Frame;
use FML\Controls\Label;
use FML\Controls\Labels\Label_Text;
use FML\Controls\Quad;
use FML\Controls\Entry;
use FML\Controls\Quads\Quad_BgsPlayerCard;
use FML\Controls\Quads\Quad_Icons128x128_1;
use FML\Controls\Quads\Quad_BgRaceScore2;
use FML\ManiaLink;
use FML\Script\Features\Paging;
use ManiaControl\Bills\BillManager;
use ManiaControl\Files\FileUtil;

class SprintRoundLiveRanking implements CallbackListener, CommandListener, ManialinkPageAnswerListener, TimerListener, Plugin {


  const ID = 97;
  const VERSION = 0.3;
  const NAME = 'ObsMania Live Round Ranking';
  const AUTHOR = 'MarlBurroW';
  const DESCRIPTION = 'Display a round live ranking widget for the sprint mode (by Guerro)';
  const WIDGET_MLID = "obsmania_sprint_roundliveranking_widget";
  const CB_PLAYERTIME = "Sprint_PlayerTime";
  const CB_PLAYERGIVEUP = "Sprint_PlayerGiveUp";
  const CB_PLAYERRESPAWN = "Sprint_PlayerRespawn";

  private $maniaControl = null;
  private $ranking = array();
  private $players = array();
  private $bestTime = null;

  private $widgetStyle = "BgRaceScore2";
  private $widgetSubstyle = "BgCardPlayer";

  private $headerStyle = "BgRaceScore2";
  private $headerSubstyle = "CupFinisher";

  private $rowStyle = "BgRaceScore2";
  private $rowSubstyle = "BgCardPlayer";



  /**
   * Return the ID of the plugin (Plugin interface)
   * @return [type] [description]
   */
  public static function getId() {
    return self::ID;
  }

  /**
   * Return the name of the plugin (Plugin interface)
   * @return [type] [description]
   */
  public static function getName() {
    return self::NAME;
  }

  /**
   * Return the version of the plugin (Plugin interface)
   * @return [type] [description]
   */
  public static function getVersion() {
    return self::VERSION;
  }

  /**
   * Return the author of the plugin (Plugin interface)
   * @return [type] [description]
   */
  public static function getAuthor() {
    return self::AUTHOR;
  }

  /**
   * Return the description of the plugin (Plugin interface)
   * @return [type] [description]
   */
  public static function getDescription() {
    return self::DESCRIPTION;
  }

  /**
   * Prepare method (Plugin interface)
   * @param  ManiaControl $mc [description]
   * @return [type]           [description]
   */
  public static function prepare(ManiaControl $mc) {

  }

  /**
   * Load the plugin (Plugin interface)
   * @param  ManiaControl $mc [description]
   * @return [type]           [description]
   */
  public function load(ManiaControl $mc) {

    // Get the Maniacontrol instance
    $this->maniaControl = $mc;

    // Settings
    $this->maniaControl->settingManager->initSetting($this, "Title", "Round Time");
    $this->maniaControl->settingManager->initSetting($this, "Position X", -138);
    $this->maniaControl->settingManager->initSetting($this, "Position Y", 35);
    $this->maniaControl->settingManager->initSetting($this, "Rows displayed", 5);

    // Common Callbacks
    $this->maniaControl->callbackManager->registerCallbackListener(SettingManager::CB_SETTING_CHANGED, $this, 'handleSettingChanged');
    $this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERCONNECT, $this, 'handlePlayerConnect');
    $this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERDISCONNECT, $this, 'handlePlayerDisconnect');
    $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::BEGINMAP, $this, 'handleBeginMap');
    $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::ENDMAP, $this, 'handleEndMap');
    $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::BEGINROUND, $this, 'handleBeginRound');
    $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::ENDROUND, $this, 'handleEndRound');
    $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::SCRIPTCALLBACK, $this, 'handleScriptCallback');
    $this->maniaControl->callbackManager->registerCallbackListener(RecordCallback::CHECKPOINT, $this, 'handleCheckpoint');
		$this->maniaControl->callbackManager->registerCallbackListener(RecordCallback::LAPFINISH, $this, 'handleLapFinish');
		$this->maniaControl->callbackManager->registerCallbackListener(RecordCallback::FINISH, $this, 'handleFinish');
    $this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERINFOCHANGED, $this, 'handlePlayerInfoChanged');

    // Sprint Script Callbacks
    $this->maniaControl->callbackManager->registerCallbackListener(self::CB_PLAYERGIVEUP, $this, 'handlePlayerGiveUp');
    //$this->maniaControl->callbackManager->registerCallbackListener(self::CB_PLAYERTIME, $this, 'handleFinish');
    $this->maniaControl->callbackManager->registerCallbackListener(self::CB_PLAYERRESPAWN, $this, 'handlePlayerRespawn');
		//$this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERRESPAWN, $this, 'handlePlayerRespawn');

    $this->displayLiveRoundRanking();

  }

  /**
   * Callback triggered when a scriptcallback is received (any), and trigger the good local callback
   * @param  [type] $name [description]
   * @param  [type] $data [description]
   * @return [type]       [description]
   */
  public function handleScriptCallback($name, $data) {

      //$this->maniaControl->chat->sendChat('CALLBACK: ' . $name);
    switch ($name) {
			case self::CB_PLAYERTIME:
				$this->maniaControl->getCallbackManager()->triggerCallback(self::CB_PLAYERTIME, $data);
				break;
      case self::CB_PLAYERGIVEUP:
				$this->maniaControl->getCallbackManager()->triggerCallback(self::CB_PLAYERGIVEUP, $data);
				break;
      case self::CB_PLAYERRESPAWN:
				$this->maniaControl->getCallbackManager()->triggerCallback(self::CB_PLAYERRESPAWN, $data);
				break;

    }
  }

  /**
   * Unload the plugin (Plugin interface)
   * @return [type] [description]
   */
  public function unload() {
    $this->hideLiveRoundRanking();
    $this->maniaControl = null;

  }



  /**
   * Playerlist change
   * @return [type]                               [description]
   */
  public function handlePlayerInfoChanged() {
    $this->displayLiveRoundRanking();
  }


  /**
   * Setting changes callback
   * @param  ManiaControlSettingsSetting $setting [description]
   * @return [type]                               [description]
   */
  public function handleSettingChanged(\ManiaControl\Settings\Setting $setting) {
    $this->displayLiveRoundRanking();
  }

  /**
   * Add the boolean attribute hasGiveUp(true) to the ranking when a player give up
   * @param  [type] $login Player who has gave up
   * @return [type]       [description]
   */
  public function handlePlayerGiveUp($login) {
		
 		 $this->updateRankingArray();
    foreach ($this->ranking as $key => $rank) {
      if($rank->playerlogin == $login){
        $this->ranking[$key]->has_gave_up = true;
      }
    }
    $this->displayLiveRoundRanking();

  }


  /**
   * Player connect callback
   * @param  Player $player [description]
   * @return [type]         [description]
   */
  public function handlePlayerConnect(Player $player){
    $this->displayLiveRoundRanking();
  }

  /**
   * Player disconnect callback
   * @param  Player $player [description]
   * @return [type]         [description]
   */
  public function handlePlayerDisconnect(Player $player){
    $this->displayLiveRoundRanking();
  }

  /**
   * Map begin callback
   * @return [type] [description]
   */
  public function handleBeginMap(){
    $this->reset();
    $this->displayLiveRoundRanking();
  }

  /**
   * Map end callback
   * @return void
   */
  public function handleEndMap(){

    $this->displayLiveRoundRanking();
  }

  /**
   * Round begin callback
   * @return void
   */
  public function handleBeginRound(){

    $this->displayLiveRoundRanking(true);
		$this->ranking = [];
  }

  /**
   * Round end callback
   * @return void
   */
  public function handleEndRound(){

    $this->displayLiveRoundRanking(false, true);
  }

  /**
   * Player finish callback
   * @return void
   */
  public function handleFinish($data){

    $this->addTimeToLiveRanking($data->player, $data->time);
    $this->updateBestTime($data->player,$data->time);

    $this->displayLiveRoundRanking();


  }

  /**
   * Checkpoint callback
   * @return void
   */
  public function handleCheckpoint($data){
		//$this->maniaControl->chat->sendChat($data );
		
  }

  /**
   * Lap finish callback
   * @return void
   */
  public function handleLapFinish(){


  }

  public function handlePlayerTime($data){

  }

  public function handlePlayerRespawn($login){
		 $this->updateRankingArray();
    foreach ($this->ranking as $key => $rank) {
      if($rank->playerlogin == $login){
			
        if(isset($this->ranking[$key]->respawns)){
          $this->ranking[$key]->respawns++;
        }else{
          $this->ranking[$key]->respawns = 1;
        }
      }
    }

    $this->displayLiveRoundRanking();
  }

  /**
   * Create a rank row for the live ranking
   * @param  string $rank     [description]
   * @param  string $nickname [description]
   * @param  string $time    [description]
   * @param  float $ypos      Y Position of the row, 0 is top
   * @return Frame            Return a Frame object
   */
  public function createRow($rank, $nickname, $time, $ypos, $oldTime = false, $has_gave_up = false, $type = 'time', $respawns = false, $finish = false){


    $rowFrame = new Frame();

    $rowFrame->setPosition(-20, $ypos, 10 );
    $rowFrame->setSize(50,5);

    $rankLabel = new Label();
    $rowFrame->add($rankLabel);
    $rankLabel->setHAlign(Control::CENTER);
    $rankLabel->setX(0);
    $rankLabel->setSize(5, 5);
    $rankLabel->setTextSize(1.5);

    if(is_int($rank)){
      switch($rank){
        case 1:
          $rankLabel->setText('$FF0' . $rank);
        break;
        case 2:
          $rankLabel->setText('$CCC' . $rank);
        break;
        case 3:
          $rankLabel->setText('$963' . $rank);
        break;
        default:
          $rankLabel->setText('$39F' . $rank);
        break;
      }
    }else{
      $rankLabel->setText($rank);
    }

    $rankLabel->setStyle('TextButtonBig');

    $nicknameLabel = new Label();
    $rowFrame->add($nicknameLabel);
    $nicknameLabel->setHAlign(Control::LEFT);
    $nicknameLabel->setX(5);
    $nicknameLabel->setSize(25, 5);
    $nicknameLabel->setTextSize(1);
    $nicknameLabel->setText($nickname);
    $nicknameLabel->setTextEmboss(true);
    $nicknameLabel->setStyle('TextButtonBig');

    $timeLabel = new Label();

    $timeLabel->setHAlign(Control::CENTER);
    $timeLabel->setX(38);
    $timeLabel->setSize(10, 5);
    $timeLabel->setTextSize(1.5);

    if($oldTime)
      $textcolor = '$999';
    else
      $textcolor = '$fff';
	
		
	
	
     

    

    if($type == 'header'){
			
      $timeLabel->setStyle('TextButtonBig');
			$timeLabel->setText('Time');
			
    }elseif($type == 'time'){
			if($time){
				$time = $this->toSMTime($time);
				$timeLabel->setText($textcolor . $time);
			}elseif($has_gave_up){
				$timeLabel->setText($textcolor . 'Gave Up');
			}elseif($respawns){
				if($oldTime || $finish){
					$timeLabel->setText($textcolor . '$nNOT FINISHED');
				}else{
					$timeLabel->setText($textcolor . $respawns . "RS");
				}
				
			}else{
				if($oldTime || $finish){
					$timeLabel->setText($textcolor . '$nNOT FINISHED');
				}else{
					$timeLabel->setText($textcolor . "RUN");
				}
				
				
			}
      $timeLabel->setStyle('TextButtonBig');
    }
		
		$rowFrame->add($timeLabel);		

    return $rowFrame;

  }

  /**
   * Hide the widget Manialink
   * @param  [type] $playerlogin [description]
   * @return void           [description]
   */
  public function hideLiveRoundRanking($playerlogin = null){
    $this->maniaControl->getManialinkManager()->hideManialink(self::WIDGET_MLID, $playerlogin);
  }


  public function displayLiveRoundRankingForPlayer(Player $player, $oldTime = false, $finish = false){

    $title = $this->maniaControl->settingManager->getSettingValue($this, "Title");
    $pos_x = $this->maniaControl->settingManager->getSettingValue($this, "Position X");
    $pos_y = $this->maniaControl->settingManager->getSettingValue($this, "Position Y");
    $rows_displayed = $this->maniaControl->settingManager->getSettingValue($this, "Rows displayed");

    $manialink = new ManiaLink(self::WIDGET_MLID);

    $size_y = 25 + ($rows_displayed * 5);

    $frame     = new Frame();
    $manialink->add($frame);

    $frame->setPosition($pos_x, $pos_y);

    $background = new Quad();

    $background->setVAlign(Control::TOP);
    $background->setStyles($this->widgetStyle, $this->widgetSubstyle);
    $frame->add($background);
    $background->setSize(50,$size_y);

    $ypos = -5;
    $rankCount = 1;

    $rowTitleFrame = new Frame();
    $rowTitleFrame->setPosition(0, $ypos);
    $timeLabel = new Label();
    $rowTitleFrame->add($timeLabel);
    $timeLabel->setHAlign(Control::CENTER);
    $timeLabel->setX(0);
    $timeLabel->setSize(50, 5);
    $timeLabel->setTextSize(2);
    $timeLabel->setText($title);
    $timeLabel->setStyle('TextButtonBig');
    $frame->add($rowTitleFrame);

    $ypos -= 4;

    $bestTimeFrame = new Frame();
    $bestTimeFrame->setPosition(0, $ypos);
    $bestTimeLabel = new Label();
    $bestTimeFrame->add($bestTimeLabel);
    $bestTimeLabel->setHAlign(Control::CENTER);
    $bestTimeLabel->setX(0);
    $bestTimeLabel->setSize(50, 5);
    $bestTimeLabel->setTextSize(0.5);
    $bestTimeLabel->setStyle('TextButtonBig');
    $bestTimeLabel->setText("MATCH BEST TIME: " . ($this->bestTime != null ? '$FF0' . $this->toSMTime($this->bestTime->time) : '-'));

    $frame->add($bestTimeFrame);

    $ypos -= 4;

    $bestTimeByFrame = new Frame();
    $bestTimeByFrame->setPosition(0, $ypos);
    $bestTimeByLabel = new Label();
    $bestTimeByFrame->add($bestTimeByLabel);
    $bestTimeByLabel->setHAlign(Control::CENTER);
    $bestTimeByLabel->setX(0);
    $bestTimeByLabel->setSize(50, 5);
    $bestTimeByLabel->setTextSize(0.5);
    $bestTimeByLabel->setStyle('TextButtonBig');
    $bestTimeByLabel->setText(($this->bestTime != null ? "By " . $this->bestTime->nickname : '-'));

    $frame->add($bestTimeByFrame);

    $ypos -= 6;

    $rowFrame = $this->createRow("#","Nickname","Time", $ypos,false,false, 'header', false, false);

    $rowBackground = new Quad();
    $rowBackground->setSize(50,5);
    $rowBackground->setX(20);
    $rowBackground->setStyles($this->headerStyle, $this->headerSubstyle);
    $rowFrame->add($rowBackground);

    $frame->add($rowFrame);
    $ypos-= 5;

    $rowcounter = 1;
    foreach ($this->ranking as $key => $rank) {
			
      if($rowcounter > $rows_displayed)
        break;

      $nickname = $rank->nickname;
			
			
			
      if ($rank->time){
        $rnk = $rankCount;
        $rankCount++;
				$time = $rank->time;
      }else{
				$time = null;
				$rnk = "-";
			}
     

      if(isset($rank->has_gave_up) && $rank->has_gave_up){
        $has_gave_up = true;
      }else {
        $has_gave_up = false;
      }

			if(isset($rank->respawns) && $rank->respawns){
        $respawns = $rank->respawns;
      }else {
        $respawns = false;
      }
			
      $rowFrame = $this->createRow($rnk,$nickname,$time, $ypos, $oldTime, $has_gave_up, 'time', $respawns, $finish);

      $rowBackground = new Quad();
      $rowBackground->setSize(50,5);
      $rowBackground->setX(20);
      $rowBackground->setStyles($this->rowStyle, $this->rowSubstyle);
      $rowFrame->add($rowBackground);

      $frame->add($rowFrame);

      $ypos-= 5;
      $rowcounter++;
    }




    $poweredByObsManiaFrame = new Frame();
    $poweredByObsManiaFrame->setPosition(0, -($size_y - 2));
    $poweredByObsManiaLabel = new Label();
    $poweredByObsManiaFrame->add($poweredByObsManiaLabel);
    $poweredByObsManiaLabel->setHAlign(Control::CENTER);
    $poweredByObsManiaLabel->setX(0);
    $poweredByObsManiaLabel->setSize(50, 4);
    $poweredByObsManiaLabel->setTextSize(0.5);
    //$poweredByObsManiaLabel->setStyle('TextSPScoreBig');
    $poweredByObsManiaLabel->setText('Plugin powered by $l[http://obsmania.com]$o$f60Obs$cccMania$l');

    $frame->add($poweredByObsManiaFrame);

    $this->maniaControl->getManialinkManager()->sendManialink($manialink, $player);

  }


  /**
   * Display the liveranking Widget
   * @param  [type] $playerlogin [description]
   * @return void             [description]
   */
  public function displayLiveRoundRanking($oldTime = false , $finish = false){

    $players = $this->maniaControl->playerManager->getPlayers();
    $this->updateRankingArray();

    foreach ($players as $player) {
      $this->displayLiveRoundRankingForPlayer($player, $oldTime , $finish);
    }

  }




  public function refreshPlayerList(){


      $players = $this->maniaControl->playerManager->getPlayers();

      if (empty($players)) {
          return array();
      }

      $playersInfo = array();

      foreach ($players as $player) {
          if($player->isSpectator) continue;
          $info = new \stdClass();
          $info->playerlogin = $player->login;
          $info->nickname = $player->nickname;

          array_push($playersInfo, $info);
      }

      $this->players = $playersInfo;


      return $playersInfo;
  }

  public function updateRankingArray(){
    $this->refreshPlayerList();


    foreach($this->players as $pkey => $player){

      $playerExistInRanking = false;

      foreach ($this->ranking as $rkey => $rank) {

        if($player->playerlogin == $rank->playerlogin){

            $playerExistInRanking = true;
            break;
        }

      }


      if(!$playerExistInRanking){

        $newRank = new \stdClass();
        $newRank->nickname = $player->nickname;
        $newRank->playerlogin = $player->playerlogin;
        $newRank->time = null;
        array_push($this->ranking, $newRank);

      }
    }

    usort($this->ranking, array($this, "orderByTimeDesc"));


  }

  public function addTimeToLiveRanking($player, $time){


    $newRank = new \stdClass();
    $newRank->playerlogin = $player->login;
    $newRank->nickname = $player->nickname;
    $newRank->time = $time;

    $exist = false;
    foreach($this->ranking as $key => $rank){
      if($rank->playerlogin == $newRank->playerlogin){
        $exist = true;
        $this->ranking[$key]->time = $newRank->time;

      }
    }
    if(!$exist){
      array_push($this->ranking, $newRank);
    }




  }

  public function toSMTime($time){

    $time         = (int)$time;
    $milliseconds = $time % 1000;
		$hundredthsseconds = $time % 10000;
    $seconds      = floor($time / 1000);
    $minutes      = floor($seconds / 60);
    $hours        = floor($minutes / 60);
    $minutes -= $hours * 60;
    $seconds -= (($hours * 60 * 60) + ($minutes * 60));
    $format = ($hours > 0 ? $hours . ':' : '');
    $format .= str_pad($minutes, 2, '0', STR_PAD_LEFT) . ":";
    $format .= str_pad($seconds, 2, '0', STR_PAD_LEFT) . ".";
		$format .= str_pad(($milliseconds / 10), 2, '0', STR_PAD_LEFT); 
    //$format .= str_pad($milliseconds, 3, '0', STR_PAD_LEFT);
    return $format;

  }

  /**
   * Sorting method that order the ranking by Rank Desc
   * @param  [type] $a [description]
   * @param  [type] $b [description]
   * @return [type]    [description]
   */
  private function orderByTimeDesc($a, $b)
  {
     if($a->time == $b->time) {
        if($a->time == '')
          return strcmp($a->playerlogin, $b->playerlogin);
        elseif($b->time == '')
          return strcmp($b->playerlogin, $a->playerlogin);
        else
          return 0;
     }
     else if($a->time == '') {
         return 1;
     }
     else if($b->time == '') {
         return -1;
     }

     return $a->time < $b->time ? -1 : 1;
  }


  private function updateBestTime($player, $time){

    if($this->bestTime == null || $time < $this->bestTime->time){
      $player->time = $time;
      $this->bestTime = $player;
      $this->maniaControl->chat->sendChat('$3f0New match best time $FF0' . $this->toSMTime($this->bestTime->time) . '$3f0 by ' .   $this->bestTime->nickname );
    }
  }


  private function reset(){
    $this->bestTime = null;
    $this->ranking = [];
    $this->players = [];
  }

}
